from unittest.mock import patch

from django.test import TestCase, override_settings
from django.test.client import Client
from django.core.urlresolvers import reverse_lazy
from django.contrib.auth.models import User

from .decorators import override_environment


example_url = reverse_lazy('example')


class TestIPRestriction(TestCase):
    """
    Test various aspects of the IpWhitelister middelware by changing Django settings.
    Other test in test_config, already test the difference between changing Django's
    settings and setting environemnt variable.  Here we just use the one method.
    """

    @classmethod
    def setUpClass(cls):
        cls.username = 'testuser'
        cls.password = '12345'
        cls.user = User.objects.create(username=cls.username)
        cls.user.set_password('12345')
        cls.user.is_superuser = True
        cls.user.save()

    @classmethod
    def tearDownClass(cls):
        cls.user.delete()

    def _get_response_code_for_ip(self, ip, url=example_url, login=False):
        # Helper function to set an originating IP address to the given IP, and request
        # our one example view, returning the response's status code
        client = Client(REMOTE_ADDR=ip)

        if login:
            client.login(username=self.user.username, password=self.password)

        resp = client.get(url)
        return resp.status_code

    def test_default_no_ip_restriction(self):
        # By default, without specifying in settings or environment variables
        # requests should be allowed and not interefed with by the middleware
        resp = self.client.get(example_url)
        self.assertEqual(resp.status_code, 200)

    @override_settings(RESTRICT_IPS=False)
    def test_no_ip_restriction(self):
        # Specifying in settings that IPs should not be restricted and confirm
        # requests are allowed and not interefed with by the middleware
        resp = self.client.get(example_url)
        self.assertEqual(resp.status_code, 200)

    @override_settings(RESTRICT_IPS=True)
    def test_basic_ip_restriction(self):
        # Enable IP restriction, but set no allowed IPs or network ranges,
        # confirm that requests return a 403
        resp = self.client.get(example_url)
        self.assertEqual(resp.status_code, 403)

    @override_settings(RESTRICT_IPS=True, ALLOWED_IPS=['127.0.0.1', '192.168.0.1'])
    def test_allowed_ips(self):
        # Restict IPs to a list of known IPs and check that they are allowed, but other IPs
        # are forbidden
        code = self._get_response_code_for_ip('127.0.0.1')
        self.assertEqual(code, 200)

        code = self._get_response_code_for_ip('192.168.0.1')
        self.assertEqual(code, 200)

        code = self._get_response_code_for_ip('127.0.0.2')
        self.assertEqual(code, 403)

        code = self._get_response_code_for_ip('192.168.0.2')
        self.assertEqual(code, 403)

    @override_settings(RESTRICT_IPS=True, ALLOWED_IP_RANGES=['192.168.0.0/31'])
    def test_allowed_ip_ranges(self):
        # Specift a very limited IP range, and check that the 2 IP within it receive a 200
        # and that an IP just beyond the range is forbidden
        code = self._get_response_code_for_ip('192.168.0.0')
        self.assertEqual(code, 200)

        code = self._get_response_code_for_ip('192.168.0.1')
        self.assertEqual(code, 200)

        code = self._get_response_code_for_ip('192.168.0.2')
        self.assertEqual(code, 403)


    @patch('ip_restriction.IpWhitelister.logger')
    @override_settings(RESTRICT_IPS=True, ALLOWED_IP_RANGES=['127.0.0.1/30'])
    def test_invalid_network_range_logging(self, mock_logger):
        # Supply a 'broken' format of CIDR ip range, check that the request is forbidden
        # and that a warning log message is produced
        code = self._get_response_code_for_ip('127.0.0.1')
        self.assertEqual(code, 403)
        self.assertTrue(mock_logger.warning.called)

    @override_settings(RESTRICT_IPS=True, ALLOW_ADMIN=True)
    def test_allow_admin(self):
        # with IP restriction on, with no allowed IPs, but ALLOWED_ADMIN=True, all views
        # be forbidden, except the admin views
        resp = self.client.get(example_url)
        self.assertEqual(resp.status_code, 403)

        # The admin views are allowed and return 200
        resp = self.client.get(reverse_lazy('admin:login'))
        self.assertEqual(resp.status_code, 200)

        # But without ALLOW_AUTHENTICATED, logging in still forbids access to the views
        code = self._get_response_code_for_ip('127.0.0.1', example_url, login=True)
        self.assertEqual(code, 403)


    @override_settings(RESTRICT_IPS=True, ALLOW_AUTHENTICATED=True)
    def test_authenticated_allowed(self):
        # Access to the views are disallowed, even the admin views, but when authenticated
        # no restrictions are in place
        resp = self.client.get(example_url)
        self.assertEqual(resp.status_code, 403)

        resp = self.client.get(reverse_lazy('admin:login'))
        self.assertEqual(resp.status_code, 403)

        code = self._get_response_code_for_ip('127.0.0.1', example_url, login=True)
        self.assertEqual(code, 200)



from django.test import TestCase, override_settings

from ip_restriction import IpWhitelister
from .decorators import override_environment


class TestIPRestrictionConfig(TestCase):

    def test_default_config(self):
        restrictor = IpWhitelister()
        self.assertFalse(restrictor.RESTRICT_IPS)
        self.assertFalse(restrictor.ALLOW_ADMIN)
        self.assertFalse(restrictor.ALLOW_AUTHENTICATED)
        self.assertEquals(restrictor.ALLOWED_IPS, [])
        self.assertEquals(restrictor.ALLOWED_IP_RANGES, [])

    @override_settings(
        RESTRICT_IPS=True,
        ALLOW_ADMIN=True,
        ALLOW_AUTHENTICATED=True,
        ALLOWED_IPS=['192.168.0.1'],
        ALLOWED_IP_RANGES=['192.168.0.0/24']
    )
    def test_django_settings(self):
        restrictor = IpWhitelister()
        self.assertTrue(restrictor.RESTRICT_IPS)
        self.assertTrue(restrictor.ALLOW_ADMIN)
        self.assertTrue(restrictor.ALLOW_AUTHENTICATED)
        self.assertEquals(restrictor.ALLOWED_IPS, ['192.168.0.1'])
        self.assertEquals(restrictor.ALLOWED_IP_RANGES, ['192.168.0.0/24'])

    @override_environment(
        RESTRICT_IPS=True,
        ALLOW_ADMIN=True,
        ALLOW_AUTHENTICATED=True,
        ALLOWED_IPS=['192.168.0.1'],
        ALLOWED_IP_RANGES=['192.168.0.0/24']
    )
    def test_django_settings(self):
        restrictor = IpWhitelister()
        self.assertTrue(restrictor.RESTRICT_IPS)
        self.assertTrue(restrictor.ALLOW_ADMIN)
        self.assertTrue(restrictor.ALLOW_AUTHENTICATED)
        self.assertEquals(restrictor.ALLOWED_IPS, ['192.168.0.1'])
        self.assertEquals(restrictor.ALLOWED_IP_RANGES, ['192.168.0.0/24'])


    @override_settings(
        RESTRICT_IPS=False,
        ALLOW_ADMIN=False,
        ALLOW_AUTHENTICATED=False,
        ALLOWED_IPS=['192.168.0.1'],
        ALLOWED_IP_RANGES=['192.168.0.0/24']
    )
    @override_environment(
        RESTRICT_IPS=True,
        ALLOW_ADMIN=True,
        ALLOW_AUTHENTICATED=True,
        ALLOWED_IPS=['192.168.0.2'],
        ALLOWED_IP_RANGES=['192.168.0.0/20']
    )
    def test_environment_trumps_settings(self):
        restrictor = IpWhitelister()
        self.assertTrue(restrictor.RESTRICT_IPS)
        self.assertTrue(restrictor.ALLOW_ADMIN)
        self.assertTrue(restrictor.ALLOW_AUTHENTICATED)
        self.assertEquals(restrictor.ALLOWED_IPS, ['192.168.0.2'])
        self.assertEquals(restrictor.ALLOWED_IP_RANGES, ['192.168.0.0/20'])



import os
import ipaddress
import logging

from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import resolve
from django.conf import settings


class IpWhitelister():
    """
    Simple middlware to allow IP addresses via settings variables ALLOWED_IPS, ALLOWED_IP_RANGES.

    Made to be compatible with Django 1.9 and also 1.10+
    """

    logger = logging.getLogger(__name__)

    def __init__(self, get_response=None):
        self.get_response = get_response
        self.RESTRICT_IPS = self._get_config_var('RESTRICT_IPS', bool)
        self.ALLOWED_IPS = self._get_config_var('ALLOWED_IPS', list)
        self.ALLOWED_IP_RANGES = self._get_config_var('ALLOWED_IP_RANGES', list)
        self.ALLOW_ADMIN = self._get_config_var('ALLOW_ADMIN', bool)
        self.ALLOW_AUTHENTICATED = self._get_config_var('ALLOW_AUTHENTICATED', bool)

    def __call__(self, request):
        response = self.process_request(request)
        
        if not response and self.get_response:
            response = self.get_response(request)
        
        return response

    def _get_config_var(self, name, vartype):
        """
        Get the whitelist config variable from the Django settings, or from the environment.
        Environment variables take preference over Django settings
        """

        env_val = os.environ.get(name)
        setting_val = getattr(settings, name, None)

        if vartype == bool:
            if env_val is None:
                return setting_val is True
            else:
                return env_val.lower() == 'true' or env_val == '1'
        else:
            if env_val is None:
                return setting_val if setting_val is not None else []
            else:
                return [val.strip() for val in env_val.split(',') if val != '']

    def get_client_ip(self, request):
        """
        Get the incoming request's originating IP, looks first for X_FORWARDED_FOR header, which is provided by some
        PaaS platforms, since the Django REMOTE_ADDR is affected by internal routing.  Fallback to the REMOTE_ADDR if
        the header is not present
        """

        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')

        if x_forwarded_for:
            ip = x_forwarded_for.split(',')[-1].strip()
        else:
            ip = request.META.get('REMOTE_ADDR')

        return ip

    def process_request(self, request):
        if self.RESTRICT_IPS:
            # Get the app name
            app_name = resolve(request.path).app_name
            authenticated = request.user.is_authenticated()

            # Allow access to the admin
            if app_name == 'admin' and self.ALLOW_ADMIN:
                return None

            # Allow access to authenticated users
            if authenticated and self.ALLOW_AUTHENTICATED:
                return None

            block_request = True
            # Get the incoming IP address
            request_ip_str = self.get_client_ip(request)
            request_ip = ipaddress.ip_address(request_ip_str)

            # If it's in the ALLOWED_IPS, don't block it
            if request_ip_str in self.ALLOWED_IPS:
                block_request = False

            # If it's within a ALLOWED_IP_RANGE, don't block it
            for allowed_range in self.ALLOWED_IP_RANGES:
                try:
                    network = ipaddress.ip_network(allowed_range)
                except ValueError as e:
                    self.logger.warning('Failed to parse specific network address: {}'.format("".join(e.args)))
                    continue

                if request_ip in network:
                    block_request = False
                    break

            # Otherwise, 403 Forbidden
            if block_request:
                raise PermissionDenied()

        return None



from os import environ


def override_environment(**envs):
    # Handy decorator, similar to override settings that will temporarily set environment
    # variables for a test, and then revert the values afterwards
    def decorator(func):
        def wrapped(*args, **kwargs):
            # Store the current environment variable values, and set new values
            curr_vals = {}
            for key, val in envs.items():
                curr_vals[key] = environ.get(key)

                # Force the given value to a string
                if type(val) == list:
                    environ[key] = ",".join(val)
                else:
                    environ[key] = "{}".format(val)

            # Run the wrapped function/method with it's args.kwargs
            try:
                func(*args, **kwargs)
            finally:
                # Restore environment variables to original values
                for key, val in curr_vals.items():
                    if val is None:
                        del environ[key]
                    else:
                        environ[key] = val

        return wrapped
    return decorator



import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = 'fake-key'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.admin',
    'tests'
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

import django
# Django changed the settings variable for middleware between 1.9 and 1.10
# Set the correct variable needed for testing based on the installed django version
version_components = [int(comp) for comp in django.get_version().split('.')]

# Basic necessary middleware for the IpWhitelister to work
_middleware = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'ip_restriction.IpWhitelister'
]

if version_components[0] == 1 and version_components[1] == 9:
    MIDDLEWARE_CLASSES = _middleware
elif version_components[0] == 1 and version_components[1] >= 10:
    MIDDLEWARE = _middleware



from django.conf.urls import url
from django.contrib import admin

from . import views


urlpatterns = [
    url(r'^example$', views.ExampleView.as_view(), name='example'),
    url(r'^admin/', admin.site.urls),
]



from django.views.generic import View
from django.core.urlresolvers import reverse_lazy
from django.http import HttpResponse


class ExampleView(View):
    def get(self, request):
        return HttpResponse("It works!")
